use std::io;

use axum::{
    routing::{get, get_service},
    extract::{Path, TypedHeader, ws},
    http::{StatusCode},
    response,
    Router};

async fn hello(Path(name): Path<String>, user_agent: Option<TypedHeader<axum::headers::UserAgent>>) -> impl response::IntoResponse {
    if let Some(TypedHeader(user_agent)) = user_agent {
        (StatusCode::OK, format!("Hello {} from {}", name, user_agent))
    } else {
        (StatusCode::OK, format!("Hello {}", name))
    }
}

async fn ws_echo(mut socket: ws::WebSocket) {
    while let Some(msg) = socket.recv().await {
        if let Ok(msg) = msg {
            tracing::debug!("websocket got message: {:?}", msg);

            let reply = match msg  {
                ws::Message::Text(t) => ws::Message::Text(t),
                ws::Message::Binary(b) => ws::Message::Binary(b),
                // axum handles Ping/Pong by itself
                ws::Message::Ping(_) => { continue },
                ws::Message::Pong(_) => { continue },
                ws::Message::Close(_) => { break }
            };

            if socket.send(reply).await
                .is_err() {
                    tracing::info!("websocket client disconnected");
                    break;
                }
        }
        else {
            tracing::info!("websocket client disconnected");
            break;
        }
    }
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt::init();
    let app = Router::new()
        .route("/hello/:name", get(hello))
        .nest("/static",
               get_service(tower_http::services::ServeDir::new(".").precompressed_gzip())
                   .handle_error(|e: io::Error| async move {
                       (StatusCode::INTERNAL_SERVER_ERROR, format!("Unhandled internal error: {}", e))
                   })
        )
        .route("/ws-echo", get(|ws: ws::WebSocketUpgrade| async {ws.on_upgrade(ws_echo)}))
        .layer(
            tower::ServiceBuilder::new()
                .layer(
                    tower_http::trace::TraceLayer::new_for_http()
                         .make_span_with(tower_http::trace::DefaultMakeSpan::default().include_headers(true)),
                )
        );

    let addr = std::net::SocketAddr::from(([127, 0, 0, 1], 3000));
    tracing::info!("listening on {}", addr);
    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await
        .unwrap();
}
